#include "interpolator.h"
#include <string.h>

static const s16 coeff[INTERPOLATOR_ORDER+1] = {
	-307,-137,-88,22,174,328,431,431,294,21,-341,-705,-953,-969,-660,8,996,2190,3421,4491,5220,5478,5220,4491,3421,2190,996,8,-660,-969,-953,-705,-341,21,294,431,431,328,174,22,-88,-137,-307
};

//==============================================

INTERPOLATOR::INTERPOLATOR(void)
{
	memset(buf_up,0,sizeof(buf_up));
	buf_up_inx = 0;
	memset(buf_down,0,sizeof(buf_down));
	buf_down_inx = 0;
	buf_down_cnt = 0;
}
//==============================================

INTERPOLATOR::~INTERPOLATOR(void)
{
}

//==============================================
static s16 clip(s32 foo)
{
	if (foo > 32767) foo = 32767;
	if (foo < -32767) foo = -32767;
	return (s16)foo;
}
//==============================================

void INTERPOLATOR::upsample(s16* buf_in, u32 size, s16* buf_out)
{
	while(size--)
	{
		buf_up[buf_up_inx] = *buf_in++;
		if (++buf_up_inx >= INTERPOLATOR_UP_SIZE) buf_up_inx = 0;
		for(u32 i = INTERPOLATOR_FACTOR; i--;)
		{
			s32 sum = 0;
			for (u32 j = 0; j < (i?INTERPOLATOR_UP_SIZE-1:INTERPOLATOR_UP_SIZE); j++)
			{
				u32 inx = buf_up_inx + j;
				if (inx >= INTERPOLATOR_UP_SIZE) inx-=INTERPOLATOR_UP_SIZE;
				sum += (s32)buf_up[inx]*coeff[INTERPOLATOR_FACTOR*j+i];
			}
			*buf_out++ = clip((sum*(s32)INTERPOLATOR_FACTOR)>>15);
			//*buf_out++ = clip(sum>>15);
		}
	}
}

//==============================================

void INTERPOLATOR::downsample(s16* buf_in, u32 size, s16* buf_out)
{
	while(size--)
	{
		buf_down[buf_down_inx] = *buf_in++;
		if (++buf_down_inx >= INTERPOLATOR_ORDER+1) buf_down_inx = 0;
		if (++buf_down_cnt >= INTERPOLATOR_FACTOR)
		{
			s32 sum = 0;
			buf_down_cnt = 0;
			for(u32 i = 0; i < INTERPOLATOR_ORDER+1; i++)
			{
				u32 inx = buf_down_inx + i;
				if (inx >= INTERPOLATOR_ORDER+1) inx-=INTERPOLATOR_ORDER+1;
				sum += (s32)buf_down[inx]*coeff[i];
			}
			*buf_out++ = clip(sum>>15);
		}
	}
}

//==============================================
