#ifndef KADEMLIA_H_INCLUDED
#define KADEMLIA_H_INCLUDED
//-----------------------------------------------------------------------
#include <windows.h>
//-----------------------------------------------------------------------
#include "types.h"
#include "params.h"
//-----------------------------------------------------------------------
typedef struct {
	u32 id[NETWORK_ID_LENGTH]; // ID of node
	u32 ip;                     // IP address
	u16 port;                   // UDP port number
} NODE;
//-----------------------------------------------------------------------
#define KAD_DHT_FLAG_PINGED		0x00000001 // node is in pinging process
#define KAD_DHT_FLAG_REPLIED	0x00000002 // node replied on ping (or sent any other messages)
#define KAD_DHT_FLAG_LOST		0x00000004 // node is lost (doesn't reply on pings)
//-----------------------------------------------------------------------
#define KAD_MAX_HASHES			4 // maximum number of hashes in user name
#define KAD_HASH_TMO			10 // 10 sec
//-----------------------------------------------------------------------
typedef struct distr_table {
	NODE node;
	u64 lastPingTime;
	u64 firstPingTime;
	u32 flags;
	u32 ping_count;
	struct distr_table *next;
} DHT;
//-----------------------------------------------------------------------
typedef struct {
	DHT* dht;
	NODE waiting_list[KADEMLIA_WAITING_MAX];
} BUCKET;
//-----------------------------------------------------------------------
class USERS;
#include "users.h"
//-----------------------------------------------------------------------
typedef struct {
	u32 hash[NETWORK_ID_LENGTH];
	u32 value[NETWORK_ID_LENGTH];
} kad_data_t;
//-----------------------------------------------------------------------
typedef struct {
	kad_data_t data;
	int timeout;
	NODE nearest[KADEMLIA_A];
	int nearest_num;
	int status;
} hashes_t;
//-----------------------------------------------------------------------
class KADEMLIA
{
	NODE me;
	HWND mainWnd;
 
	CRITICAL_SECTION cs;
	BUCKET grp[NETWORK_ID_LENGTH*32]; // DHT tables for each bucket

	HANDLE sysThread;
	HANDLE eStopThreads;
	friend DWORD WINAPI systemProc(KADEMLIA* kad);

	HANDLE eFoundNode;
	bool inFinding;
	bool inBootstrapping;
	DHT findingDHT[KADEMLIA_K];
	NODE findigNode;

	kad_data_t *data;
	u32 data_size;

	hashes_t my_hashes[KAD_MAX_HASHES];
	int my_hashes_size;

	hashes_t search_hashes[KAD_MAX_HASHES];
	int search_hashes_size;

	void sendPing(const NODE* nd, bool reply);
	void sendRequestID(const NODE* nd, const u32 *id, bool isData);
	void sendReplyID(const NODE* nd, const u32 *id, u32 cmd);
	void requestID(void);
	u32 fillBucket(const u32* id, u32* buf);
	void findindReply(const u32* pack, u32 pack_len, const NODE* nd);
	void placeHash(hashes_t* h);
	void hashTimeout(hashes_t* h);
	void requestHash(int num);
	void CheckBootstrapping(void);

public:
	KADEMLIA(const NODE *my_node, HWND mWnd);
	~KADEMLIA();

	void start();
	void stop();
    bool isStarted(){return (sysThread!=NULL);}
    void packetReceived(u32* packet, u32 size, const NODE* nd, USERS* usr);
	void updateNode(const NODE* nd, bool trueReply, USERS* usr);

	DHT* get_first_node(u32 bucket) {return grp[bucket].dht;};
	DHT* get_next_node(DHT* cur){return cur->next;};
 
	bool findNode(const u32 *id, NODE *nd);
	void findNode(LPTSTR name);
	bool findNode(u32 ip, NODE *nd);
	void pingNode(const NODE *nd);
};
//-----------------------------------------------------------------------
bool isIdIdentical(const u32* id1, const u32* id2);
//-----------------------------------------------------------------------
bool isIdNonZero(const u32* id);
//-----------------------------------------------------------------------
#endif //KADEMLIA_H_INCLUDED
