#include <winsock2.h>
#include <Mswsock.h>
#include "network.h"
#include "password.h"
#include "resource.h"
#include "log.h"
#include "hash_func.h"
#include "random.h"
#include "diffie_helman.h"
#include "main.h"
#include <tchar.h>
#include <CommCtrl.h>
#include <Windows.h>
//-----------------------------------------
static LPTSTR module = TEXT("Network");
//-----------------------------------------
CONNECTION::CONNECTION(const NODE* node)
{
	CopyMemory(&nd,node,sizeof(NODE));
	flags = 0;
	next_connection = NULL;
	est = NULL;
	pack_num_est = 0;
	pack_num_sec = 0;
	pack_num_drop = 0;
	pack_num_sent = 0;
	timeout = 0;
	away = false;
	cp_strcchcpy((tchar_t*)userInfo,sizeof(userInfo)/sizeof(tchar_t),_T("<unknown>"));
	userInfoSize = (u32)cp_strlen((tchar_t*)userInfo)*sizeof(tchar_t);
	rx_sign = NULL;
	tx_sign = NULL;
	conf_sign = NULL;
	InitializeCriticalSection(&est_cs);
}
//-----------------------------------------
CONNECTION::~CONNECTION()
{
	if (rx_sign) WaitForSingleObject(rx_sign,INFINITE);
	if (tx_sign) WaitForSingleObject(tx_sign,INFINITE);
	if (conf_sign) WaitForSingleObject(conf_sign,INFINITE);
	EnterCriticalSection(&est_cs);
	if (est)
	{
		est->tx_session.reset();
		est->rx_session.reset();
		delete est;
	}
	est = NULL;
	LeaveCriticalSection(&est_cs);
	DeleteCriticalSection(&est_cs);
}
//-----------------------------------------------------------------------
DWORD WINAPI receiverProc(NETWORK* n);
DWORD WINAPI timeoutProc(NETWORK* n);
//-----------------------------------------------------------------------
NETWORK::NETWORK(NODE* me, HWND mWnd)
{
	InitializeCriticalSection(&cs);
	con = NULL;
	away = false;

	CopyMemory(my_id,me->id,NETWORK_ID_LENGTH*sizeof(u32));

    mainWnd = mWnd;
	recThread = NULL;

	// open new socket
	soc = socket(AF_INET,SOCK_DGRAM,IPPROTO_UDP);
	if (soc == INVALID_SOCKET)
	{
		MessageBox(mWnd,TEXT("Can't create socket"),TEXT("Error"),MB_OK|MB_ICONERROR);
		return;
	}
	ZeroMemory(&my_addr,sizeof(my_addr));
	my_addr.sin_family = AF_INET;
	my_addr.sin_addr.s_addr = INADDR_ANY;
	my_addr.sin_port = htons(me->port);
	if (SOCKET_ERROR == bind(soc,(sockaddr*)&my_addr,sizeof(my_addr)))
	{
		MessageBox(mWnd,TEXT("Can't bind socket to local port"),TEXT("Error"),MB_OK|MB_ICONERROR);
		return;
	}
    // set receiver timeout in ms
	DWORD timeout = 100;
	setsockopt(soc,SOL_SOCKET,SO_RCVTIMEO,(char*)&timeout,sizeof(timeout));
	// disable  new behavior using
	// IOCTL: SIO_UDP_CONNRESET
	DWORD dwBytesReturned = 0;
	BOOL bNewBehavior = FALSE;
	WSAIoctl(soc, SIO_UDP_CONNRESET,&bNewBehavior, sizeof(bNewBehavior),NULL,0,&dwBytesReturned,NULL,NULL);
	// Initialize sending packet
    u8 passw[password_max_length];
	u32 pas_len = getPassword(passw);
	if (pas_len)
	{
		sending_size = cmn_tx_session.init_session((u8*)sending_buff, getSecretData(), passw, pas_len);
		sending_size_1 = cmn_tx_session_1.init_session_1((u8*)sending_buff_1);
		// create receiver thread and stop event
		eStopThreads = CreateEvent(NULL,TRUE,FALSE,NULL);
		recThread = CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)receiverProc,this,0,NULL);
		tmoThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)timeoutProc, this, 0, NULL);
	}
	else
	{
		recThread = NULL;
	}
}
//-----------------------------------------------------------------------
NETWORK::~NETWORK()
{
	u32 buffer_coded[100];
	sockaddr_in addr = {0};
	addr.sin_family = AF_INET;
	CopyMemory(buffer_coded,my_id,NETWORK_ID_LENGTH*sizeof(u32));
	buffer_coded[NETWORK_ID_LENGTH] = SESSION_TERMINATE|(away?SESSION_USER_AWAY:0);
	if (recThread)
	{
		SetEvent(eStopThreads);
		WaitForSingleObject(recThread,INFINITE);
		CloseHandle(recThread);
		WaitForSingleObject(tmoThread,INFINITE);
		CloseHandle(tmoThread);
	}
	if (con) // clear all connections
	{
		while(con)
		{
			CONNECTION* cn = con->next_connection;
			addr.sin_addr.s_addr = htonl(con->nd.ip);
			addr.sin_port = htons(con->nd.port);
			LogLowLevel(true,con->nd.ip,con->nd.port,con->nd.id,buffer_coded[NETWORK_ID_LENGTH]&0xFF,(NETWORK_ID_LENGTH+1)*sizeof(u32));
			sendto(soc,(char*)buffer_coded,(NETWORK_ID_LENGTH+1)*sizeof(u32),0,(sockaddr*)&addr,sizeof(sockaddr_in));
			delete con;
			con = cn;
		}
	}
	closesocket(soc);
    DeleteCriticalSection(&cs);
	cmn_tx_session.reset();
	cmn_tx_session_1.reset();
}
//-----------------------------------------------------------------------
CONNECTION* NETWORK::find_connection(const NODE* nd, bool establish)
{
	CONNECTION *s, *prev_s;
	if (!con) // there is no any session
	{
		if (establish)
		{
			con = new CONNECTION(nd);
			return con;
		}
		else
		{
			return NULL;
		}
	}
	for (s = con; s; s = s->next_connection)
	{
		if (isIdIdentical(s->nd.id,nd->id))
		{
			return s;
		}
		prev_s = s;
	}
	if (!establish) return NULL;
	prev_s->next_connection = new CONNECTION(nd);
	return prev_s->next_connection;
}
//-----------------------------------------------------------------------
void NETWORK::send_big_buffer(const NODE* nd, u32* buf, u32 size, u32 command, u32* pack_cnt)
{
	u32 tx_buffer[NETWORK_SIZE_UDP/sizeof(u32)];
	sockaddr_in addr = {0};
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = htonl(nd->ip);
	addr.sin_port = htons(nd->port);
	
	CopyMemory(tx_buffer,my_id,NETWORK_ID_LENGTH*sizeof(u32));

	u32 total_pack = (size+NETWORK_SIZE_SECURED-1)/NETWORK_SIZE_SECURED; // number of packets for sending
	for (u32 p = 0; p < total_pack; p++)
	{
		u32 chunk_size = (p==(total_pack-1)) ? (size-(total_pack-1)*NETWORK_SIZE_SECURED) : NETWORK_SIZE_SECURED; // size of data chunk
		tx_buffer[NETWORK_ID_LENGTH] = command|(away?SESSION_USER_AWAY:0)|(total_pack<<8)|(p<<16);
		CopyMemory(&tx_buffer[NETWORK_ID_LENGTH+1],&buf[p*NETWORK_SIZE_SECURED/sizeof(u32)],chunk_size);
		LogLowLevel(true,nd->ip,nd->port,nd->id,command,chunk_size+(NETWORK_ID_LENGTH+1)*sizeof(u32));
		sendto(soc,(char*)tx_buffer,chunk_size+(NETWORK_ID_LENGTH+1)*sizeof(u32),0,(sockaddr*)&addr,sizeof(sockaddr_in));
	}
	*pack_cnt += total_pack;
}
//-----------------------------------------------------------------------
bool NETWORK::collect_big_buffer(u32* cur_buf, u32 size, ESTABLISHMENT* est)
{
	u32 total_pack = (cur_buf[0]>>8)&0xFF;
	u32 pack_num = (cur_buf[0]>>16)&0xFF;
	if ((pack_num >= total_pack) || (total_pack*NETWORK_SIZE_SECURED > NETWORK_SIZE_DH_MAX) ||
		((pack_num != (total_pack-1)) && (size != (NETWORK_SIZE_SECURED+4))) || 
		((pack_num == (total_pack-1)) && (size >= (NETWORK_SIZE_SECURED+4)))) // error
	{
		est->rec_flags = 0;
		return false;
	}
	CopyMemory(&est->buff[pack_num*NETWORK_SIZE_SECURED/sizeof(u32)],&cur_buf[1],size-4);
	est->rec_flags |= 1<<pack_num;
	if (pack_num == (total_pack-1)) // last packet
	{
		est->rec_size = (total_pack-1)*NETWORK_SIZE_SECURED + size-4;
	}
	for (u32 i = 0; i < total_pack; i++)
	{
		if (!(est->rec_flags & (1<<i))) return false;
	}
	return true;
}
//-----------------------------------------------------------------------
u32 NETWORK::send_packet(PACKET_FOR_SEND* pack)
{
	u32 size = 0;
	u32 tx_buffer_plain[2048];
	u32 tx_buffer_coded[2048];
	CONNECTION* s;
	sockaddr_in addr = {0};
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = htonl(pack->nd->ip);
	addr.sin_port = htons(pack->nd->port);

	if (isIdIdentical(pack->nd->id,my_id)) return 0; // don't send data to myself

	CopyMemory(tx_buffer_coded,my_id,NETWORK_ID_LENGTH*sizeof(u32));

    if (!isIdNonZero(pack->nd->id)) // for zero ID just send ping
    {
		EnterCriticalSection(&cs);
		for (s = con; s; s = s->next_connection)
		{
			if ((s->nd.ip == pack->nd->ip) && (s->nd.port == pack->nd->port))
			{
				memcpy((void*)pack->nd->id,s->nd.id,NETWORK_ID_LENGTH*sizeof(u32));
				break;
			}
		}
		LeaveCriticalSection(&cs);
		if (!s)
		{
			tx_buffer_coded[NETWORK_ID_LENGTH] = SESSION_PING_REQ|(away?SESSION_USER_AWAY:0);
			tx_buffer_coded[NETWORK_ID_LENGTH+1] = NETWORK_PROTOCOL_VERSION;
			LogLowLevel(true,pack->nd->ip,pack->nd->port,pack->nd->id,tx_buffer_coded[NETWORK_ID_LENGTH]&0xFF,(NETWORK_ID_LENGTH+2)*sizeof(u32));
			sendto(soc,(char*)tx_buffer_coded,(NETWORK_ID_LENGTH+2)*sizeof(u32),0,(sockaddr*)&addr,sizeof(sockaddr_in));
			return 0;
		}
    }

	EnterCriticalSection(&cs);
	s = find_connection(pack->nd,true);

	if (!s) // not enough memory
	{
		LeaveCriticalSection(&cs);
		return 0;
	}
	
	if (s->flags&SESSION_REJECTION && !acceptOldProtocol())
	{
		LeaveCriticalSection(&cs);
		return 0;
	}

	if(s->flags&SESSION_LAUNCHED)
	{
		addr.sin_addr.s_addr = htonl(s->nd.ip);
		addr.sin_port = htons(s->nd.port);
		tx_buffer_plain[0] = pack->protocol;
		CopyMemory(&tx_buffer_plain[1], pack->buffer, pack->buf_len);
		tx_buffer_coded[NETWORK_ID_LENGTH] = SESSION_ENCODED|(away?SESSION_USER_AWAY:0);
		size = s->ses.encrypt((u8*)&tx_buffer_coded[NETWORK_ID_LENGTH+1],(u8*)tx_buffer_plain,pack->buf_len+sizeof(u32));
		LogLowLevel(true,s->nd.ip,s->nd.port,s->nd.id,tx_buffer_coded[NETWORK_ID_LENGTH]&0xFF,size+(NETWORK_ID_LENGTH+1)*sizeof(u32));
        size = (u32)sendto(soc,(char*)tx_buffer_coded,size+(NETWORK_ID_LENGTH+1)*sizeof(u32),0,(sockaddr*)&addr,sizeof(sockaddr_in));
		s->pack_num_sent++;
		s->flags++;
		if ((s->flags&SESSION_REQ_COUNT_MASK) >= SESSION_REQ_COUNT_MAX) // so long time no any answer. Reset session
		{
			s->flags = 0;
			s->ses.reset();
		}
		LeaveCriticalSection(&cs);
	}
	else
	{
		if(!(s->flags&SESSION_PINGED) || !(s->flags&SESSION_RESPONDED)) // if node isn't pinged or didn't respond, send ping request command
		{
			tx_buffer_coded[NETWORK_ID_LENGTH] = SESSION_PING_REQ|(away?SESSION_USER_AWAY:0);
			tx_buffer_coded[NETWORK_ID_LENGTH+1] = NETWORK_PROTOCOL_VERSION;
			s->flags |= SESSION_PINGED|SESSION_ESTABLISHING;
			LogLowLevel(true,pack->nd->ip,pack->nd->port,pack->nd->id,tx_buffer_coded[NETWORK_ID_LENGTH]&0xFF,(NETWORK_ID_LENGTH+2)*sizeof(u32));
			sendto(soc,(char*)tx_buffer_coded,(NETWORK_ID_LENGTH+2)*sizeof(u32),0,(sockaddr*)&addr,sizeof(sockaddr_in));
			s->pack_num_sent++;
			LeaveCriticalSection(&cs);
		}
		else if(s->est) // pinged and responded, but didn't send answer yet. Resend request again
		{
			if (!s->timeout && !s->tx_sign && !s->rx_sign && !s->conf_sign)
			{
				TCHAR str[400];
				s->timeout = 50;
				if (s->flags&SESSION_FIRST_STAGE)
				{
					tx_buffer_coded[NETWORK_ID_LENGTH] = SESSION_NOT_CONFIRMED|(away?SESSION_USER_AWAY:0);
					LogLowLevel(true,pack->nd->ip,pack->nd->port,pack->nd->id,tx_buffer_coded[NETWORK_ID_LENGTH]&0xFF,(NETWORK_ID_LENGTH+1)*sizeof(u32));
					sendto(soc,(char*)tx_buffer_coded,(NETWORK_ID_LENGTH+1)*sizeof(u32),0,(sockaddr*)&addr,sizeof(sockaddr_in));
					s->pack_num_sent++;
					_stprintf_s(str, 200, TEXT("Request confirmation from %d.%d.%d.%d:%d"), FIRST_IPADDRESS(pack->nd->ip), SECOND_IPADDRESS(pack->nd->ip), THIRD_IPADDRESS(pack->nd->ip), FOURTH_IPADDRESS(pack->nd->ip), pack->nd->port);
				}
				else if (s->flags&SESSION_ESTABLISHING)
				{
					NODE ndd;
					CopyMemory(ndd.id,s->nd.id,NETWORK_ID_LENGTH*sizeof(u32));
					ndd.ip = s->est->ip;
					ndd.port = s->est->port;
					send_big_buffer(&ndd, s->est->sent_buf, s->est->sent_size, SESSION_REQUEST, &s->pack_num_sent);
					_stprintf_s(str, 200, TEXT("Again send Session request to %d.%d.%d.%d:%d"), FIRST_IPADDRESS(pack->nd->ip), SECOND_IPADDRESS(pack->nd->ip), THIRD_IPADDRESS(pack->nd->ip), FOURTH_IPADDRESS(pack->nd->ip), pack->nd->port);
				}
				else
				{
					s->flags = 0;
					_stprintf_s(str, 200, TEXT("Timeout for connection from %d.%d.%d.%d:%d"), FIRST_IPADDRESS(pack->nd->ip), SECOND_IPADDRESS(pack->nd->ip), THIRD_IPADDRESS(pack->nd->ip), FOURTH_IPADDRESS(pack->nd->ip), pack->nd->port);
				}
				LeaveCriticalSection(&cs);
				LogMessage(module, str);
			}
			else
			{
				LeaveCriticalSection(&cs);
			}
		}
		else // wrong condition. Pinged, responded but didn't create establishment
		{
			s->flags = 0;
			s->ses.reset();
			LeaveCriticalSection(&cs);
		}
	}
	return size;
}
//-----------------------------------------------------------------------
s8 name_compare(const u8* buf_ref,const u8* buf_test,u32 max_size)
{
	while (max_size--)
	{
		s8 diff = *buf_ref++ - *buf_test++;
		if (diff) return diff;
	}
	return 0;
}
//-----------------------------------------------------------------------
DWORD WINAPI timeoutProc(NETWORK* n)
{
	CONNECTION* s;

	while (WAIT_TIMEOUT == WaitForSingleObject(n->eStopThreads, 100))
	{
		EnterCriticalSection(&n->cs);
		for (s = n->con; s; s = s->next_connection)
		{
			if (s->timeout) s->timeout--;
		}
		LeaveCriticalSection(&n->cs);
	}
	return 0;
}
//-----------------------------------------------------------------------
DWORD WINAPI rx_signature(CONNECTION* s)
{
    u8 passw[password_max_length];
	u32 pas_len = getPassword(passw);
	u32 session_id[fingerprint_length_x32];
	NETWORK* n = getNet();
	TCHAR str[400];
	EnterCriticalSection(&s->est_cs);
	if (s->est)
	{
		if (s->est->protocol >= 2)
		{
			s->est->sent_size = s->est->rx_session.response_to_init_session_1((u8*)s->est->sent_buf,(u8*)s->est->buff, getSecretData(), passw, pas_len);
			s->userInfoSize = 0;
		}
		else
		{
			s->est->sent_size = s->est->rx_session.response_to_init_session((u8*)s->est->sent_buf,(u8*)s->est->buff, getSecretData(), passw, pas_len);
			if ((s->est->sent_size > 0) && s->est->rx_session.session_is_active(session_id) && isIdIdentical(s->nd.id,session_id))
			{
				s->userInfoSize = s->est->rx_session.get_user_info(s->userInfo);
			}
			else
			{
				s->userInfoSize = 0;
				s->est->sent_size = 0;
			}
		}
		s->userInfo[s->userInfoSize] = s->userInfo[s->userInfoSize+1] = 0;
		if ((s->est->sent_size > 0))
		{
			n->send_big_buffer(&s->nd,s->est->sent_buf,s->est->sent_size,SESSION_RESPONSE,&s->pack_num_sent);
			if (s->est->protocol <= 1)
			{
				s->est->rx_session.start_initialized_session(&s->ses);
				s->flags = SESSION_LAUNCHED|SESSION_JUST_SETUP;
				s->nd.ip = s->est->ip;
				s->nd.port = s->est->port;
				_stprintf_s(str,200,TEXT("Successful Session request from %d.%d.%d.%d:%d, user name: %s, fingerprint %08X %08X %08X %08X %08X %08X %08X %08X"),FIRST_IPADDRESS(s->nd.ip),SECOND_IPADDRESS(s->nd.ip),THIRD_IPADDRESS(s->nd.ip),FOURTH_IPADDRESS(s->nd.ip),s->nd.port,
					s->userInfo,session_id[0],session_id[1],session_id[2],session_id[3],session_id[4],session_id[5],session_id[6],session_id[7]);
			}
			else
			{
				s->flags |= SESSION_FIRST_STAGE;
				_stprintf_s(str,200,TEXT("Wait confirmation because of successful session request from %d.%d.%d.%d:%d"),FIRST_IPADDRESS(s->est->ip),SECOND_IPADDRESS(s->est->ip),THIRD_IPADDRESS(s->est->ip),FOURTH_IPADDRESS(s->est->ip),s->est->port);
			}
			s->est->last_size = s->est->rec_size;
			memcpy(s->est->last_buf,s->est->buff,s->est->rec_size);
			s->est->rec_size = 0;
			s->est->rec_flags = 0;
		}
		else
		{
			_stprintf_s(str,200,TEXT("Wrong Session request from %d.%d.%d.%d:%d"),FIRST_IPADDRESS(s->est->ip),SECOND_IPADDRESS(s->est->ip),THIRD_IPADDRESS(s->est->ip),FOURTH_IPADDRESS(s->est->ip),s->est->port);
			s->flags &= ~(SESSION_PINGED|SESSION_RESPONDED|SESSION_ESTABLISHING|SESSION_FIRST_STAGE);
			s->est->tx_session.reset();
			s->est->rx_session.reset();
			delete s->est;
			s->est = NULL;
			// WARNING: someone sent incorrect request
		}
		LeaveCriticalSection(&s->est_cs);
		LogMessage(module,str);
		if (s->est && (s->est->protocol != NETWORK_PROTOCOL_VERSION))
		{
			_stprintf_s(str,200,TEXT("Another protocol version on %d.%d.%d.%d:%d"),FIRST_IPADDRESS(s->est->ip),SECOND_IPADDRESS(s->est->ip),THIRD_IPADDRESS(s->est->ip),FOURTH_IPADDRESS(s->est->ip),s->est->port);
			LogMessage(module,str);
		}
	}
	else LeaveCriticalSection(&s->est_cs);
	CloseHandle(s->rx_sign);
	s->rx_sign = NULL;
	return 0;
}
//-----------------------------------------------------------------------
DWORD WINAPI tx_signature(CONNECTION* s)
{
    u8 passw[password_max_length];
	u32 pas_len = getPassword(passw);
	u32 session_id[fingerprint_length_x32];
	NETWORK* n = getNet();
	TCHAR str[400];
	EnterCriticalSection(&s->est_cs);
	if (s->est)
	{
		if (s->est->protocol >= 2)
		{
			s->est->conf_size = s->est->tx_session.process_response_confirm_session_1((u8*)s->est->conf_buf,(u8*)s->est->buff,getSecretData(),passw,pas_len);
			if ((s->est->conf_size > 0) && s->est->tx_session.session_is_active_1(session_id))
			{
				s->userInfoSize = s->est->tx_session.get_user_info_1(s->userInfo);
			}
			else
			{
				s->est->conf_size = 0;
				s->userInfoSize = 0;
			}
		}
		else
		{
			s->est->conf_size = s->est->tx_session.process_response((u8*)s->est->buff);
			if ((s->est->conf_size > 0) && s->est->tx_session.session_is_active(session_id))
			{
				s->userInfoSize = s->est->tx_session.get_user_info(s->userInfo);
			}
			else
			{
				s->est->conf_size = 0;
				s->userInfoSize = 0;
			}
		}
		s->userInfo[s->userInfoSize] = s->userInfo[s->userInfoSize+1] = 0;
		if ((s->est->conf_size > 0) && isIdIdentical(s->nd.id,session_id)) // normal response
		{
			s->est->tx_session.start_initialized_session(&s->ses);
			s->flags = SESSION_LAUNCHED|SESSION_JUST_SETUP;
			s->nd.ip = s->est->ip;
			s->nd.port = s->est->port;
			_stprintf_s(str,200,TEXT("Successful Session response from %d.%d.%d.%d:%d, user name: %s, fingerprint %08X %08X %08X %08X %08X %08X %08X %08X"),FIRST_IPADDRESS(s->nd.ip),SECOND_IPADDRESS(s->nd.ip),THIRD_IPADDRESS(s->nd.ip),FOURTH_IPADDRESS(s->nd.ip),s->nd.port,
				s->userInfo,session_id[0],session_id[1],session_id[2],session_id[3],session_id[4],session_id[5],session_id[6],session_id[7]);
			if (s->est->protocol >= 2)
			{
				n->send_big_buffer(&s->nd,s->est->conf_buf,s->est->conf_size,SESSION_CONFIRMATION,&s->pack_num_sent);
			}
			s->est->last_size = s->est->rec_size;
			memcpy(s->est->last_buf,s->est->buff,s->est->rec_size);
			s->est->rec_size = 0;
			s->est->rec_flags = 0;
		}
		else
		{
			_stprintf_s(str,200,TEXT("Wrong Session response from %d.%d.%d.%d:%d"),FIRST_IPADDRESS(s->est->ip),SECOND_IPADDRESS(s->est->ip),THIRD_IPADDRESS(s->est->ip),FOURTH_IPADDRESS(s->est->ip),s->est->port);
			s->flags &= ~(SESSION_PINGED|SESSION_RESPONDED|SESSION_ESTABLISHING|SESSION_FIRST_STAGE);
			s->est->tx_session.reset();
			s->est->rx_session.reset();
			delete s->est;
			s->est = NULL;
			// WARNING: Someone use ID which is not equal to fingerpint
		}
		LeaveCriticalSection(&s->est_cs);
		LogMessage(module,str);
	} else LeaveCriticalSection(&s->est_cs);
	CloseHandle(s->tx_sign);
	s->tx_sign = NULL;
	return 0;
}
//-----------------------------------------------------------------------
DWORD WINAPI conf_signature(CONNECTION* s)
{
	TCHAR str[400];
	u32 session_id[fingerprint_length_x32];
	NETWORK* n = getNet();
	EnterCriticalSection(&s->est_cs);
	if (s->est)
	{
		u32 size = s->est->rx_session.check_confirmation_1((u8*)s->est->buff);
		if ((size > 0) && s->est->rx_session.session_is_active_1(session_id) && isIdIdentical(s->nd.id,session_id)) // normal response
		{
			s->est->rx_session.start_initialized_session(&s->ses);
			s->userInfoSize = s->est->rx_session.get_user_info_1(s->userInfo);
			s->userInfo[s->userInfoSize] = s->userInfo[s->userInfoSize+1] = 0;
			s->flags = SESSION_LAUNCHED|SESSION_JUST_SETUP;
			s->nd.ip = s->est->ip;
			s->nd.port = s->est->port;
			_stprintf_s(str,200,TEXT("Successful Session Confirmation from %d.%d.%d.%d:%d, user name: %s, fingerprint %08X %08X %08X %08X %08X %08X %08X %08X"),FIRST_IPADDRESS(s->nd.ip),SECOND_IPADDRESS(s->nd.ip),THIRD_IPADDRESS(s->nd.ip),FOURTH_IPADDRESS(s->nd.ip),s->nd.port,
				s->userInfo,session_id[0],session_id[1],session_id[2],session_id[3],session_id[4],session_id[5],session_id[6],session_id[7]);
			s->est->last_size = s->est->rec_size;
			memcpy(s->est->last_buf,s->est->buff,s->est->rec_size);
			s->est->rec_size = 0;
			s->est->rec_flags = 0;
		}
		else
		{
			_stprintf_s(str,200,TEXT("Wrong Session confirmation from %d.%d.%d.%d:%d"),FIRST_IPADDRESS(s->est->ip),SECOND_IPADDRESS(s->est->ip),THIRD_IPADDRESS(s->est->ip),FOURTH_IPADDRESS(s->est->ip),s->est->port);
			s->flags &= ~(SESSION_PINGED|SESSION_RESPONDED|SESSION_ESTABLISHING|SESSION_FIRST_STAGE);
			s->est->tx_session.reset();
			s->est->rx_session.reset();
			delete s->est;
			s->est = NULL;
		}
		LeaveCriticalSection(&s->est_cs);
		LogMessage(module,str);
	} else LeaveCriticalSection(&s->est_cs);
	CloseHandle(s->conf_sign);
	s->conf_sign = NULL;
	return 0;
}
//-----------------------------------------------------------------------
DWORD WINAPI receiverProc(NETWORK* n)
{
	TCHAR str[400];
	int pack_len;
	sockaddr_in addr = {0};
	CONNECTION* s;
	u32 size;
	u32 buffer_plain[2048];
	u32 buffer_coded[2048];
	int fault_cnt = 10;

    while (WAIT_TIMEOUT == WaitForSingleObject(n->eStopThreads,0))
	{
		int addr_size = sizeof(addr);
		memset(&addr,0,sizeof(addr));
		addr.sin_family = AF_INET;
		pack_len = recvfrom(n->soc,(char*)buffer_coded,sizeof(buffer_coded),0,(sockaddr*)&addr,&addr_size);
		KADEMLIA* kad = getKad();
		if (pack_len == SOCKET_ERROR) // error parsing
        {
			int err;
			if ((err = WSAGetLastError()) == WSAETIMEDOUT) continue;
			if (!--fault_cnt) 
			{
				MessageBox(n->mainWnd,TEXT("To many attempts to open socket\nProgram will be closed"),TEXT("Error"),MB_OK|MB_ICONERROR);
				PostMessage(n->mainWnd,WM_CLOSE,0,0);
				return 0;
			}
			//MessageBox(n->mainWnd,TEXT("Socket error occures\nReopen socket"),TEXT("Error"),MB_OK|MB_ICONERROR);
			closesocket(n->soc);
			// open new socket
			n->soc = socket(AF_INET,SOCK_DGRAM,IPPROTO_UDP);
			if (n->soc == INVALID_SOCKET)
			{
				MessageBox(n->mainWnd,TEXT("Can't create socket"),TEXT("Error"),MB_OK|MB_ICONERROR);
				return 0;
			}
			if (SOCKET_ERROR == bind(n->soc,(sockaddr*)&n->my_addr,sizeof(n->my_addr)))
			{
				MessageBox(n->mainWnd,TEXT("Can't bind socket to local port"),TEXT("Error"),MB_OK|MB_ICONERROR);
				return 0;
			}
			// disable  new behavior using
			// IOCTL: SIO_UDP_CONNRESET
			DWORD dwBytesReturned = 0;
			BOOL bNewBehavior = FALSE;
			WSAIoctl(n->soc, SIO_UDP_CONNRESET,&bNewBehavior, sizeof(bNewBehavior),NULL,0,&dwBytesReturned,NULL,NULL);
			// set receiver timeout in ms
			DWORD timeout = 100;
			setsockopt(n->soc,SOL_SOCKET,SO_RCVTIMEO,(char*)&timeout,sizeof(timeout));
			continue;
        }
        if (pack_len >= (int)(NETWORK_ID_LENGTH+1)*sizeof(u32))
        {
            NODE nd;
			nd.ip = ntohl(addr.sin_addr.s_addr);
			nd.port = ntohs(addr.sin_port);
            CopyMemory(nd.id,buffer_coded,NETWORK_ID_LENGTH*sizeof(u32));
			if (isIdIdentical(nd.id,n->my_id)) continue; // don't process data from myself (something wrong with received packet)
            EnterCriticalSection(&n->cs);
            s = n->find_connection(&nd,true);
			u8 pack_type = buffer_coded[NETWORK_ID_LENGTH]&0xFF;
			s->away = (buffer_coded[NETWORK_ID_LENGTH]&SESSION_USER_AWAY)!=0;
			LogLowLevel(false,nd.ip,nd.port,nd.id,pack_type,pack_len);
            switch(pack_type)
            {
            case SESSION_REQUEST:
				if (s->rx_sign || s->tx_sign || s->conf_sign || ((s->flags&SESSION_ESTABLISHING) && (nd.id[0] > n->my_id[0]) && s->timeout) || !s->est)
				{
				    LeaveCriticalSection(&n->cs);
					break; // we are in the process of signing this connection
				}
				else if (!(s->flags&(SESSION_PINGED|SESSION_RESPONDED))) // it's not correct to run session without ping handshake
				{
					CopyMemory(buffer_coded,n->my_id,NETWORK_ID_LENGTH*sizeof(u32));
					buffer_coded[NETWORK_ID_LENGTH] = SESSION_TERMINATE|(n->away?SESSION_USER_AWAY:0);
					LogLowLevel(true,nd.ip,nd.port,nd.id,buffer_coded[NETWORK_ID_LENGTH]&0xFF,(NETWORK_ID_LENGTH+1)*sizeof(u32));
					sendto(n->soc,(char*)buffer_coded,(NETWORK_ID_LENGTH+1)*sizeof(u32),0,(sockaddr*)&addr,sizeof(sockaddr_in));
				    LeaveCriticalSection(&n->cs);
					break; // we are in the process of signing this connection
				}
				s->pack_num_est++;
				s->flags &= ~SESSION_ESTABLISHING; // make sure to kill out establishing
				if (TryEnterCriticalSection(&s->est_cs))
				{
					if (n->collect_big_buffer(&buffer_coded[NETWORK_ID_LENGTH],pack_len-NETWORK_ID_LENGTH*sizeof(u32),s->est))
					{
						bool same_request = false;
						if (s->est->last_size && s->est->sent_size && (s->est->last_size == s->est->rec_size))
						{
							same_request = true;
							for (u32 i = 0; i < s->est->rec_size/4; i++)
								if (s->est->buff[i] != s->est->last_buf[i])
								{
									same_request = false;
									break;
								}
						}
						if (same_request)
						{
							n->send_big_buffer(&nd,s->est->sent_buf,s->est->sent_size,SESSION_RESPONSE,&s->pack_num_sent);
							s->est->rec_size = 0;
							s->est->rec_flags = 0;
							LeaveCriticalSection(&s->est_cs);
							LeaveCriticalSection(&n->cs);
							_stprintf_s(str,200,TEXT("Same Session request from %d.%d.%d.%d:%d"),FIRST_IPADDRESS(nd.ip),SECOND_IPADDRESS(nd.ip),THIRD_IPADDRESS(nd.ip),FOURTH_IPADDRESS(nd.ip),nd.port);
							LogMessage(module,str);
							break;
						}
						else
						{
							s->rx_sign = CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)rx_signature,s,0,NULL);
						}
					}
					LeaveCriticalSection(&s->est_cs);
				}
			    LeaveCriticalSection(&n->cs);
                break;
            case SESSION_RESPONSE:
				if (!s->tx_sign && !s->rx_sign && !s->conf_sign && (s->flags&SESSION_PINGED) && (s->flags&SESSION_RESPONDED) && (s->flags&SESSION_ESTABLISHING)) // we are in the process of signing this connection
				{
					s->pack_num_est++;
					if(TryEnterCriticalSection(&s->est_cs))
					{
						if (s->est)
						{
							if (n->collect_big_buffer(&buffer_coded[NETWORK_ID_LENGTH],pack_len-NETWORK_ID_LENGTH*sizeof(u32),s->est))
							{
								s->tx_sign = CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)tx_signature,s,0,NULL);
							}
						}
						LeaveCriticalSection(&s->est_cs);
					}
				}
				else if (!(s->flags&(SESSION_PINGED|SESSION_RESPONDED))) // it's not correct to run session without ping handshake
				{
					CopyMemory(buffer_coded,n->my_id,NETWORK_ID_LENGTH*sizeof(u32));
					buffer_coded[NETWORK_ID_LENGTH] = SESSION_TERMINATE|(n->away?SESSION_USER_AWAY:0);
					LogLowLevel(true,nd.ip,nd.port,nd.id,buffer_coded[NETWORK_ID_LENGTH]&0xFF,(NETWORK_ID_LENGTH+1)*sizeof(u32));
					sendto(n->soc,(char*)buffer_coded,(NETWORK_ID_LENGTH+1)*sizeof(u32),0,(sockaddr*)&addr,sizeof(sockaddr_in));
				}
                LeaveCriticalSection(&n->cs);
                break;
            case SESSION_CONFIRMATION:
				if (!s->tx_sign && !s->rx_sign && !s->conf_sign && !(s->flags&SESSION_ESTABLISHING) && (s->flags&SESSION_PINGED) && (s->flags&SESSION_RESPONDED) && (s->flags&SESSION_FIRST_STAGE)) // we are in the process of signing this connection
				{
					s->pack_num_est++;
					if(TryEnterCriticalSection(&s->est_cs))
					{
						if (s->est)
						{
							if (n->collect_big_buffer(&buffer_coded[NETWORK_ID_LENGTH],pack_len-NETWORK_ID_LENGTH*sizeof(u32),s->est))
							{
								s->conf_sign = CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)conf_signature,s,0,NULL);
							}
						}
						LeaveCriticalSection(&s->est_cs);
					}
				}
                LeaveCriticalSection(&n->cs);
                break;
            case SESSION_NOT_CONFIRMED:
				s->pack_num_est++;
				if(TryEnterCriticalSection(&s->est_cs))
				{
					if ((s->flags&SESSION_LAUNCHED) && s->est && !s->timeout)
					{
						n->send_big_buffer(&nd,s->est->conf_buf,s->est->conf_size,SESSION_CONFIRMATION,&s->pack_num_sent);
						s->timeout = 50;
						_stprintf_s(str,200,TEXT("Again send session confirmation to %d.%d.%d.%d:%d"),FIRST_IPADDRESS(nd.ip),SECOND_IPADDRESS(nd.ip),THIRD_IPADDRESS(nd.ip),FOURTH_IPADDRESS(nd.ip),nd.port);
						LeaveCriticalSection(&s->est_cs);
						LeaveCriticalSection(&n->cs);
						LogMessage(module,str);
						break;
					}
					LeaveCriticalSection(&s->est_cs);
				}
	            LeaveCriticalSection(&n->cs);
                break;
            case SESSION_TERMINATE:
			case SESSION_NOT_ESTABLISHED:
				s->pack_num_est++;
				if ((s->nd.ip == nd.ip) && (s->nd.port == nd.port) && !(s->flags&SESSION_JUST_SETUP) && TryEnterCriticalSection(&s->est_cs))
				{
					if (s->est && !s->tx_sign && !s->rx_sign && !s->conf_sign) // delete wrong session
					{
						s->est->tx_session.reset();
						s->est->rx_session.reset();
						delete s->est;
						s->est = NULL;
					}
					LeaveCriticalSection(&s->est_cs);
					s->flags = 0;
					s->ses.reset();
					LeaveCriticalSection(&n->cs);
					_stprintf_s(str,200,TEXT("Received termination request from %d.%d.%d.%d:%d"),FIRST_IPADDRESS(nd.ip),SECOND_IPADDRESS(nd.ip),THIRD_IPADDRESS(nd.ip),FOURTH_IPADDRESS(nd.ip),nd.port);
					LogMessage(module,str);
					break;
				}
	            LeaveCriticalSection(&n->cs);
                break;
            case SESSION_PING_REQ:	// just send response on this request
				s->pack_num_est++;
				if (!(s->flags&SESSION_ESTABLISHING) || (nd.id[0] > n->my_id[0]) || !s->timeout)
				{
					s->flags |= SESSION_PINGED | SESSION_RESPONDED;
					s->flags &= ~(SESSION_ESTABLISHING|SESSION_FIRST_STAGE|SESSION_JUST_SETUP|SESSION_REJECTION);
					if(TryEnterCriticalSection(&s->est_cs))
					{
						if (!s->est) s->est = new ESTABLISHMENT;
						s->est->rec_flags = 0;
						s->est->rec_size = 0;
						s->est->last_size = 0;
						s->est->sent_size = 0;
						s->est->conf_size = 0;
						s->est->ip = nd.ip;
						s->est->port = nd.port;
						s->est->protocol = (pack_len < (int)(NETWORK_ID_LENGTH+2)*sizeof(u32)) ? 1 : buffer_coded[NETWORK_ID_LENGTH+1];
						if(s->est->protocol <= 1) s->est->tx_session = n->cmn_tx_session;
						else s->est->tx_session = n->cmn_tx_session_1;
						s->timeout = 500;
						LeaveCriticalSection(&s->est_cs);
					}
					CopyMemory(buffer_coded,n->my_id,NETWORK_ID_LENGTH*sizeof(u32));
					if (s->est && ((s->est->protocol >= 2) || acceptOldProtocol()))
					{
						buffer_coded[NETWORK_ID_LENGTH] = SESSION_PING_RESP|(n->away?SESSION_USER_AWAY:0);
						buffer_coded[NETWORK_ID_LENGTH+1] = NETWORK_PROTOCOL_VERSION;
						LogLowLevel(true,nd.ip,nd.port,nd.id,buffer_coded[NETWORK_ID_LENGTH]&0xFF,(NETWORK_ID_LENGTH+2)*sizeof(u32));
						sendto(n->soc,(char*)buffer_coded,(NETWORK_ID_LENGTH+2)*sizeof(u32),0,(sockaddr*)&addr,sizeof(sockaddr_in));
						s->pack_num_sent++;
						LeaveCriticalSection(&n->cs);
					}
					else
					{
						LeaveCriticalSection(&n->cs);
						_stprintf_s(str,400,TEXT("Rejected ping from %d.%d.%d.%d:%d because of old protocol"),FIRST_IPADDRESS(nd.ip),SECOND_IPADDRESS(nd.ip),THIRD_IPADDRESS(nd.ip),FOURTH_IPADDRESS(nd.ip),nd.port);
						LogMessage(module,str);
					}
				}
				else LeaveCriticalSection(&n->cs);
                break;
            case SESSION_PING_RESP: // answer on our request
				s->pack_num_est++;
                if ((s->flags&SESSION_PINGED) && !(s->flags&SESSION_RESPONDED) && (s->flags&SESSION_ESTABLISHING))
                {
                    s->flags |= SESSION_RESPONDED;
					s->flags &= ~SESSION_FIRST_STAGE;
					if(TryEnterCriticalSection(&s->est_cs))
					{
						if (!s->est) s->est = new ESTABLISHMENT;
						s->est->rec_flags = 0;
						s->est->rx_session.reset();
						s->est->rec_size = 0;
						s->est->ip = nd.ip;
						s->est->port = nd.port;
						s->timeout = 500;
						s->est->protocol = (pack_len < (int)(NETWORK_ID_LENGTH+2)*sizeof(u32)) ? 1 : buffer_coded[NETWORK_ID_LENGTH+1];
						if(s->est->protocol <= 1) s->est->tx_session = n->cmn_tx_session;
						else s->est->tx_session = n->cmn_tx_session_1;
						LeaveCriticalSection(&s->est_cs);
					}
					if (s->est && ((s->est->protocol >= 2) || acceptOldProtocol()))
					{
						if (s->est->protocol < 2)
						{
							s->est->sent_size = n->sending_size;
							CopyMemory(s->est->sent_buf,n->sending_buff,n->sending_size);
						}
						else
						{
							s->est->sent_size = n->sending_size_1;
							CopyMemory(s->est->sent_buf,n->sending_buff_1,n->sending_size_1);
						}
						n->send_big_buffer(&nd,s->est->sent_buf,s->est->sent_size,SESSION_REQUEST,&s->pack_num_sent);
						_stprintf_s(str,200,TEXT("Send Session request to %d.%d.%d.%d:%d"),FIRST_IPADDRESS(nd.ip),SECOND_IPADDRESS(nd.ip),THIRD_IPADDRESS(nd.ip),FOURTH_IPADDRESS(nd.ip),nd.port);
					}
					else
					{
						s->flags &= ~(SESSION_PINGED|SESSION_RESPONDED|SESSION_ESTABLISHING|SESSION_FIRST_STAGE);
						s->flags |= SESSION_REJECTION;
						if(TryEnterCriticalSection(&s->est_cs))
						{
							if (s->est)
							{
								s->est->tx_session.reset();
								s->est->rx_session.reset();
								delete s->est;
								s->est = NULL;
							}
							LeaveCriticalSection(&s->est_cs);
						}
						_stprintf_s(str,200,TEXT("Another protocol version on %d.%d.%d.%d:%d, connection is not approved"),FIRST_IPADDRESS(nd.ip),SECOND_IPADDRESS(nd.ip),THIRD_IPADDRESS(nd.ip),FOURTH_IPADDRESS(nd.ip),nd.port);
					}
                    LeaveCriticalSection(&n->cs);
					LogMessage(module,str);
					break;
                }
                else LeaveCriticalSection(&n->cs);
                break;
            case SESSION_ENCODED:
			case SESSION_NAME_REQ:
			case SESSION_NAME_RESP:
                if (s->flags & SESSION_LAUNCHED)
                {
                    u32 decr_size = s->ses.decrypt((u8*)buffer_plain,(u8*)&buffer_coded[NETWORK_ID_LENGTH+1],pack_len-(NETWORK_ID_LENGTH+1)*sizeof(u32));
                    if (decr_size > sizeof(u32))
                    {
	                    s->flags = SESSION_LAUNCHED; // clear all flags and counter, because responded normally
						if (kad) kad->updateNode(&nd,true,getUserList());
						if (TryEnterCriticalSection(&s->est_cs))
						{
							if (s->est && !s->tx_sign && !s->rx_sign && !s->conf_sign) // session established normaly, don't need data anymore
							{
								s->est->tx_session.reset();
								s->est->rx_session.reset();
								delete s->est;
								s->est = NULL;
							}
							LeaveCriticalSection(&s->est_cs);
						}
						s->pack_num_sec++;
						if (pack_type == SESSION_ENCODED)
						{
							NET_PACKET np;
							np.buffer = buffer_plain;
							np.buf_len = decr_size;
							np.nd = &nd;
							LeaveCriticalSection(&n->cs);
							if (WAIT_TIMEOUT == WaitForSingleObject(n->eStopThreads,0))
							{
								parsePacket(&np);
							}
							break;
						}
						else if (pack_type == SESSION_NAME_REQ)
						{
							CONNECTION* co;
							if (decr_size >= buffer_plain[0])
							{
								for (co = n->con; co; co = co->next_connection)
								{
									if (!name_compare((u8*)&buffer_plain[1],co->userInfo,min(co->userInfoSize,buffer_plain[0])))
									{
										CopyMemory(buffer_coded,n->my_id,NETWORK_ID_LENGTH*sizeof(u32));
										buffer_coded[NETWORK_ID_LENGTH] = SESSION_NAME_RESP|(n->away?SESSION_USER_AWAY:0);
										CopyMemory(buffer_plain,co->nd.id,NETWORK_ID_LENGTH*sizeof(u32));
										size = s->ses.encrypt((u8*)&buffer_coded[NETWORK_ID_LENGTH+1],(u8*)buffer_plain,NETWORK_ID_LENGTH*sizeof(u32));
										LogLowLevel(true,nd.ip,nd.port,nd.id,buffer_coded[NETWORK_ID_LENGTH]&0xFF,(NETWORK_ID_LENGTH+1)*sizeof(u32)+size);
										sendto(n->soc,(char*)buffer_coded,(NETWORK_ID_LENGTH+1)*sizeof(u32)+size,0,(sockaddr*)&addr,sizeof(sockaddr_in));
									}
								}
							}
						}
						else
						{
							LeaveCriticalSection(&n->cs);
							if (decr_size >= NETWORK_ID_LENGTH*sizeof(u32))
							{
								u32* id = new u32[NETWORK_ID_LENGTH];
								CopyMemory(id,buffer_plain,NETWORK_ID_LENGTH*sizeof(u32));
								PostMessage(n->mainWnd,WM_USER_FOUND,0,(LPARAM)id);
							}
							break;
						}
                    }
                    else
					{
						s->pack_num_drop++;
					}
                }
                else if (s->flags & SESSION_FIRST_STAGE)
				{
					CopyMemory(buffer_coded,n->my_id,NETWORK_ID_LENGTH*sizeof(u32));
					buffer_coded[NETWORK_ID_LENGTH] = SESSION_NOT_CONFIRMED|(n->away?SESSION_USER_AWAY:0);
					LogLowLevel(true,nd.ip,nd.port,nd.id,buffer_coded[NETWORK_ID_LENGTH]&0xFF,(NETWORK_ID_LENGTH+1)*sizeof(u32));
					sendto(n->soc,(char*)buffer_coded,(NETWORK_ID_LENGTH+1)*sizeof(u32),0,(sockaddr*)&addr,sizeof(sockaddr_in));
					s->pack_num_sent++;
				}
                else if (s->flags & (SESSION_PINGED|SESSION_RESPONDED))
				{
					CopyMemory(buffer_coded,n->my_id,NETWORK_ID_LENGTH*sizeof(u32));
					buffer_coded[NETWORK_ID_LENGTH] = SESSION_NOT_ESTABLISHED|(n->away?SESSION_USER_AWAY:0);
					LogLowLevel(true,nd.ip,nd.port,nd.id,buffer_coded[NETWORK_ID_LENGTH]&0xFF,(NETWORK_ID_LENGTH+1)*sizeof(u32));
					sendto(n->soc,(char*)buffer_coded,(NETWORK_ID_LENGTH+1)*sizeof(u32),0,(sockaddr*)&addr,sizeof(sockaddr_in));
					s->pack_num_sent++;
				}
				else
				{
					CopyMemory(buffer_coded,n->my_id,NETWORK_ID_LENGTH*sizeof(u32));
					buffer_coded[NETWORK_ID_LENGTH] = SESSION_TERMINATE|(n->away?SESSION_USER_AWAY:0);
					LogLowLevel(true,nd.ip,nd.port,nd.id,buffer_coded[NETWORK_ID_LENGTH]&0xFF,(NETWORK_ID_LENGTH+1)*sizeof(u32));
					sendto(n->soc,(char*)buffer_coded,(NETWORK_ID_LENGTH+1)*sizeof(u32),0,(sockaddr*)&addr,sizeof(sockaddr_in));
					s->pack_num_sent++;
					s->pack_num_drop++;
				}
				LeaveCriticalSection(&n->cs);
                break;
            }
        }
	}
	return 0;
}
//---------------------------------------------------
CONNECTION* NETWORK::get_next_node(CONNECTION* cur)
{
    CONNECTION* curc = cur;
    if (cur)
    {
        return cur->next_connection;
    }
    return cur;
}
//---------------------------------------------------
void NETWORK::get_counters(const NODE* nd, u32* est, u32* sec, u32* drop, u32* sent, u32* missed)
{
	*est = 0;
	*sec = 0;
	*drop = 0;
	*sent = 0;
	*missed = 0;
    EnterCriticalSection(&cs);
	CONNECTION* s = find_connection(nd,false);
	if (s)
	{
		*est = s->pack_num_est;
		*sec = s->pack_num_sec;
		*drop = s->pack_num_drop;
		*sent = s->pack_num_sent;
		*missed = (u32)s->ses.dropped_rx_bytes();
	}
	LeaveCriticalSection(&cs);
}
//---------------------------------------------------
void NETWORK::get_user_info(NODE* nd, LPTSTR info)
{
    EnterCriticalSection(&cs);
	CONNECTION* s = find_connection(nd,false);
	if (s)
	{
		memcpy(info,s->userInfo,s->userInfoSize);
		info[s->userInfoSize/sizeof(TCHAR)]=0;
	}
	else
	{
		cp_strcchcpy(info,user_info_max_length_bytes/sizeof(tchar_t),_T("<unknown>"));
	}
	LeaveCriticalSection(&cs);
}
//---------------------------------------------------
bool NETWORK::findUserByName(u8* userInfo, u16 size, u32* id)
{
	CONNECTION *s;
	u32 tx_buffer_plain[200];
	u32 tx_buffer_coded[220];
	sockaddr_in addr = {0};
	//
	if (size > user_info_max_length_bytes) return false;
    EnterCriticalSection(&cs);
	// try to find locally
	for (s = con; s; s = s->next_connection)
	{
		if (!name_compare(userInfo,s->userInfo,min(s->userInfoSize,size)))
		{
			u32* id_msg = new u32[NETWORK_ID_LENGTH];
			memcpy(id,s->nd.id,sizeof(u32)*NETWORK_ID_LENGTH);
			memcpy(id_msg,s->nd.id,sizeof(u32)*NETWORK_ID_LENGTH);
			LeaveCriticalSection(&cs);
			PostMessage(mainWnd,WM_USER_FOUND,0,(LPARAM)id_msg);
			return true;
		}
	}
	// try to find externally
	CopyMemory(tx_buffer_coded,my_id,NETWORK_ID_LENGTH*sizeof(u32));
	for (s = con; s; s = s->next_connection)
	{
		if(s->flags&SESSION_LAUNCHED)
		{
			addr.sin_family = AF_INET;
			addr.sin_addr.s_addr = htonl(s->nd.ip);
			addr.sin_port = htons(s->nd.port);
			tx_buffer_plain[0] = size;
			CopyMemory(&tx_buffer_plain[1], userInfo, size);
			tx_buffer_coded[NETWORK_ID_LENGTH] = SESSION_NAME_REQ|(away?SESSION_USER_AWAY:0);
			size = s->ses.encrypt((u8*)&tx_buffer_coded[NETWORK_ID_LENGTH+1],(u8*)tx_buffer_plain,((size+3)/4)*4+sizeof(u32));
			LogLowLevel(true,s->nd.ip,s->nd.port,s->nd.id,tx_buffer_coded[NETWORK_ID_LENGTH]&0xFF,size+(NETWORK_ID_LENGTH+1)*sizeof(u32));
			size = (u32)sendto(soc,(char*)tx_buffer_coded,size+(NETWORK_ID_LENGTH+1)*sizeof(u32),0,(sockaddr*)&addr,sizeof(sockaddr_in));
			s->pack_num_sent++;
			s->flags++;
			if ((s->flags&SESSION_REQ_COUNT_MASK) >= SESSION_REQ_COUNT_MAX) // so long time no any answer. Reset session
			{
				s->flags = 0;
				s->ses.reset();
			}
		}	
	}
	LeaveCriticalSection(&cs);
	return false;
}
//---------------------------------------------------
int NETWORK::get_user_status(const u32* id)
{
	CONNECTION *s;
	int res = USER_STATUS_OFFLINE;
	//
    EnterCriticalSection(&cs);
	for (s = con; s; s = s->next_connection)
	{
		if (isIdIdentical(id,s->nd.id))
		{
			if (s->flags&(SESSION_LAUNCHED|SESSION_RESPONDED)) res = s->away?USER_STATUS_AWAY:USER_STATUS_ONLINE;
			break;
		}
	}
	LeaveCriticalSection(&cs);
	return res;
}
//---------------------------------------------------
bool NETWORK::findNode(const u32 *id, NODE *nd)
{
	CONNECTION *s;
	bool res = false;
	//
    EnterCriticalSection(&cs);
	for (s = con; s; s = s->next_connection)
	{
		if (isIdIdentical(id,s->nd.id))
		{
			*nd = s->nd;
			res = true;
			break;
		}
	}
	LeaveCriticalSection(&cs);
	return res;
}
//---------------------------------------------------
void NETWORK::update_ip(NODE* nd)
{
	CONNECTION *s;
	bool res = false;
	//
    EnterCriticalSection(&cs);
	for (s = con; s; s = s->next_connection)
	{
		if (isIdIdentical(nd->id,s->nd.id))
		{
			nd->ip = s->nd.ip;
			nd->port = s->nd.port;
			break;
		}
	}
	LeaveCriticalSection(&cs);
}
//---------------------------------------------------
void NETWORK::reset_sessions(void)
{
	CONNECTION *s;
    u8 passw[password_max_length];
	u32 pas_len = getPassword(passw);
	if (pas_len)
	{
		sending_size = cmn_tx_session.init_session((u8*)sending_buff, getSecretData(), passw, pas_len);
		sending_size_1 = cmn_tx_session_1.init_session_1((u8*)sending_buff_1);
	}
    EnterCriticalSection(&cs);
	for (s = con; s; s = s->next_connection)
	{
		s->flags = 0;
		s->ses.reset();
	}
	LeaveCriticalSection(&cs);
}
//---------------------------------------------------
